/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#include <stdio.h>
#include <math.h>

#include "misc.h"
#include "ss.h"
#include "bpred.h"

struct bpred *
bpred_create(enum bpred_class class, unsigned long size)
{
  struct bpred *pred;

  if (!(pred = calloc(1, sizeof(struct bpred))))
    fatal("out of virtual memory");

  pred->class = class;

  switch (class) {
  case BPred2bitBTB:
    if (!size || (size & (size-1)) != 0)
      fatal("BTB size, `%d', must be non-zero and a power of two", size);
    pred->data.btb.size = size;
    if (!(pred->data.btb.sets = calloc(size, sizeof(struct bpred_btb_ent))))
      fatal("cannot allocate BTB storage");
    break;
  case BPredTaken:
  case BPredNotTaken:
    break;
  default:
    panic("bogus predictor class");
  }

  return pred;
}

/* print branch predictor configuration */
void
bpred_config(struct bpred *pred, FILE *stream)
{
  switch (pred->class) {
  case BPred2bitBTB:
    fprintf(stream, "pred: btb - 2 bit: %ld entries, direct-mapped\n",
	    pred->data.btb.size);
    break;
  case BPredTaken:
    fprintf(stream, "pred: predict taken\n");
    break;
  case BPredNotTaken:
    fprintf(stream, "pred: predict not taken\n");
    break;
  default:
    panic("bogus branch predictor class");
  }
}

/* print predictor stats */
void
bpred_stats(struct bpred *pred, FILE *stream)
{
  fprintf(stream, "pred: prediction rate = %f\n",
	  (double)pred->hits/(double)(pred->hits+pred->misses));
}

/* return TARGET if address is currently recorded as taken */

#define BTB_HASH(PRED, ADDR)						\
  ((((ADDR) >> 16)^(ADDR)) & ((PRED)->data.btb.size-1))

SS_ADDR_TYPE
bpred_lookup(struct bpred *pred, SS_ADDR_TYPE baddr, SS_ADDR_TYPE btarget)
{
  switch (pred->class) {
  case BPred2bitBTB:
    {
      struct bpred_btb_ent *p = &pred->data.btb.sets[BTB_HASH(pred, baddr)];
      if (p->addr == baddr)
	return ((p->state >= 2)
		? /* taken */ p->target
		: /* not taken */ 0);
      /* no record of this address holding a control inst */
      return /* not taken */ 0;
    }
    break;
  case BPredTaken:
    return btarget;
  case BPredNotTaken:
    return baddr + sizeof(SS_INST_TYPE);
  default:
    panic("bogus predictor class");
  }
}

void
bpred_update(struct bpred *pred, SS_ADDR_TYPE baddr,
	     SS_ADDR_TYPE btarget, int taken, int correct)
{
  /* update stats */
  if (correct)
    pred->hits++;
  else
    pred->misses++;

  /* update state */
  switch (pred->class) {
  case BPred2bitBTB:
    {
      struct bpred_btb_ent *p = &pred->data.btb.sets[BTB_HASH(pred, baddr)];
      if (p->addr == baddr)
	{
	  /* update current information */
	  if (taken)
	    {
	      if (p->state < 3)
		p->state++;
	      p->target = btarget;
	    }
	  else
	    {
	      /* not taken */
	      if (p->state > 0)
		p->state--;
	      /* leave target alone */
	    }      
	}
      else
	{
	  /* enter a new branch in the table */
	  p->addr = baddr;
	  p->state = (taken) ? 2 : 1;
	  p->target = btarget;
	}
    }  
    break;
  case BPredTaken:
  case BPredNotTaken:
    /* no state */
    break;
  default:
    panic("bogus predictor class");
  }
}  

void
bpred_dump(struct bpred *pred, FILE *stream)
{
  int i;

  if (!stream)
    stream = stderr;

  fprintf(stream, "Branch predictor: type: %s\n",
	  pred->class == BPred2bitBTB ? "btb w/ 2-bit saturating counters"
	  : pred->class == BPredTaken ? "static predict taken"
	  : pred->class == BPredNotTaken ? "static predict not taken"
	  : (abort(), ""));

  switch (pred->class) {
  case BPred2bitBTB:
    fprintf(stderr, "@@@ BTB contents:\n");
    for (i=0; i<pred->data.btb.size; i++)
      if (pred->data.btb.sets[i].addr)
	fprintf(stream,
		"\t[%5d]  addr: 0x%08x  state: %3d %s  target: 0x%08x\n",
		(int)i, (int)pred->data.btb.sets[i].addr,
		(int)pred->data.btb.sets[i].state,
		pred->data.btb.sets[i].state >= 2 ? "(taken)" : "(not taken)",
		(int)pred->data.btb.sets[i].target);
      else
	fprintf(stream, "\t[%5d]  unused\n", (int)i);
    fprintf(stderr, "@@@ end of BTB contents\n");
    break;
  case BPredTaken:
  case BPredNotTaken:
    /* no state */
    break;
  default:
    panic("bogus predictor class");
  }
}

