/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#include <stdio.h>
#include <stdlib.h>
#include "misc.h"
#include "bfd.h"
#include "ss.h"
#include "regs.h"
#include "memory.h"
#include "sim.h"
#include "loader.h"

/* text segment base address */
SS_ADDR_TYPE ld_text_base = 0;

/* text segment byte size */
SS_ADDR_TYPE ld_text_size = 0;

/* data segment (uninitialized + initialized) base address */
SS_ADDR_TYPE ld_data_base = 0;

/* data segment (uninitialized + initialized) byte size */
SS_ADDR_TYPE ld_data_size = 0;

/* stack segment bottom (highest) address */
SS_ADDR_TYPE ld_stack_base = SS_STACK_BASE;

/* stack segment byte size */
SS_ADDR_TYPE ld_stack_size = 0;

/* program entry point address */
SS_ADDR_TYPE ld_prog_entry = 0;

/* program environment base pointer, is initial SP value */
SS_ADDR_TYPE ld_environ_base = 0;

/* non-zero if target file is big endian */
int ld_target_big_endian;

/* copy args and environ into the simulator's MAIN MEMORY, the format in
   memory is:

   --> higher memory

   SS_STACK_BASE -> [unused]
                    [envp strings]
		    [argv strings]
		    [envp array,NULL]
                    [argv array,NULL]
   initial sp ->    [argc]

   --> lower memory
*/
void
ld_load_prog(mem_access_fn mem_fn,
	     int argc, char **argv, char **envp,
	     int zero_bss_segs)
{
  bfd *abfd;
  asection *sect;
  SS_ADDR_TYPE sp, data_break = 0, null_ptr = 0, argv_addr, envp_addr;
  SS_WORD_TYPE temp;
  int i;

  /* set up a local stack pointer, this is where the argv and envp
     data is written into program memory */
  ld_stack_base = SS_STACK_BASE;
  sp = ROUND_DOWN(SS_STACK_BASE - SS_MAX_ENVIRON, sizeof(SS_DOUBLE_TYPE));
  ld_stack_size = ld_stack_base - sp;

  /* initial stack pointer value */
  ld_environ_base = sp;

  /* load the program into memory */
  if (!(abfd = bfd_openr(argv[0], "ss-coff-big")))
    if (!(abfd = bfd_openr(argv[0], "ss-coff-little")))
      fatal("cannot open executable `%s'", argv[0]);

  /* this call is mainly for its side effect of reading in the sections.
     we follow the traditional behavior of `strings' in that we don't
     complain if we don't recognize a file to be an object file.  */
  if (!bfd_check_format(abfd, bfd_object))
    {
      bfd_close(abfd);
      fatal("cannot open executable `%s'", argv[0]);
    }

  /* record endian of target */
  ld_target_big_endian = abfd->xvec->byteorder_big_p;

  debug("processing %d sections in `%s'...",
	bfd_count_sections(abfd), argv[0]);

  for (sect=abfd->sections; sect; sect=sect->next)
    {
      char *p;

      debug("processing section `%s', %d bytes @ 0x%08x...",
	    bfd_section_name(abfd, sect), bfd_section_size(abfd, sect),
	    bfd_section_vma(abfd, sect));

      /* read the section data */
      if ((bfd_get_section_flags(abfd, sect) & SEC_ALLOC)
	  && (bfd_get_section_flags(abfd, sect) & SEC_LOAD)
	  && bfd_section_vma(abfd, sect)
	  && bfd_section_size(abfd, sect))
	{
	  if (!(p = calloc(bfd_section_size(abfd, sect), sizeof(char))))
	    fatal("cannot allocate %d bytes for section `%s'",
		  bfd_section_size(abfd, sect), bfd_section_name(abfd, sect));

	  if (!bfd_get_section_contents(abfd, sect, p, (file_ptr)0,
					bfd_section_size(abfd, sect)))
	    fatal("could not read entire `%s' section from executable",
		  bfd_section_name(abfd, sect));

	  /* copy it into simulator memory */
	  mem_bcopy(mem_fn, Write, bfd_section_vma(abfd, sect),
		    p, bfd_section_size(abfd, sect));
	  free(p);
	}
      else if (zero_bss_segs
	       && (bfd_get_section_flags(abfd, sect) & SEC_LOAD)
	       && bfd_section_vma(abfd, sect)
	       && bfd_section_size(abfd, sect))
	{
	  /* zero out the section region */
	  mem_bzero(mem_fn,
		    bfd_section_vma(abfd, sect), bfd_section_size(abfd, sect));
	}
      /* else: do nothing, is or will be init'ed to zero at page fault time */

      /* expected text section */
      if (!strcmp(bfd_section_name(abfd, sect), ".text"))
	{
	  /* .text section processing */
	  ld_text_size =
	    ((bfd_section_vma(abfd, sect) + bfd_section_size(abfd, sect))
	     - SS_TEXT_BASE) + /* for speculative fetches/decodes */128;
	}
      /* expected data sections */
      else if (!strcmp(bfd_section_name(abfd, sect), ".rdata")
	       || !strcmp(bfd_section_name(abfd, sect), ".data")
	       || !strcmp(bfd_section_name(abfd, sect), ".sdata")
	       || !strcmp(bfd_section_name(abfd, sect), ".bss")
	       || !strcmp(bfd_section_name(abfd, sect), ".sbss"))
	{
	  /* data section processing */
	  if (bfd_section_vma(abfd, sect) + bfd_section_size(abfd, sect) >
	      data_break)
	    data_break = (bfd_section_vma(abfd, sect) +
			  bfd_section_size(abfd, sect));
	}
      else
	fatal("encountered unknown section `%s', %d bytes @ 0x%08x",
	      bfd_section_name(abfd, sect), bfd_section_size(abfd, sect),
	      bfd_section_vma(abfd, sect));
    }

  /* compute data segment size from data break point */
  ld_text_base = SS_TEXT_BASE;
  ld_data_base = SS_DATA_BASE;
  ld_prog_entry = bfd_get_start_address(abfd);
  ld_data_size = data_break - ld_data_base;

  if (!bfd_close(abfd))
    fatal("could not close executable `%s'", argv[0]);

  if (!ld_text_base || !ld_text_size)
    fatal("executable is missing a `.text' section");
  if (!ld_data_base || !ld_data_size)
    fatal("executable is missing a `.data' section");
  if (!ld_prog_entry)
    fatal("program entry point not specified");

  /* determine byte/words swapping required to execute on this host */
  sim_swap_bytes = (endian_host_byte_order() != endian_target_byte_order());
  if (sim_swap_bytes)
    fprintf(stderr, "sim: *WARNING*: swapping bytes to match host...\n");
  sim_swap_words = (endian_host_word_order() != endian_target_word_order());
  if (sim_swap_words)
    fprintf(stderr, "sim: *WARNING*: swapping words to match host...\n");

  /* write [argc] */
  temp = SWAP_WORD(argc);
  mem_fn(Write, sp, &temp, sizeof(SS_WORD_TYPE));
  sp += sizeof(SS_WORD_TYPE);

  /* skip past argv array and NULL */
  argv_addr = sp;
  sp = sp + (argc + 1) * sizeof(SS_PTR_TYPE);

  /* save space for envp array and NULL */
  envp_addr = sp;
  for (i=0; envp[i]; i++)
    sp += sizeof(SS_PTR_TYPE);
  sp += sizeof(SS_PTR_TYPE);

  /* fill in the argv pointer array and data */
  for (i=0; i<argc; i++)
    {
      /* write the argv pointer array entry */
      temp = SWAP_WORD(sp);
      mem_fn(Write, argv_addr + i*sizeof(SS_PTR_TYPE),
	     &temp, sizeof(SS_PTR_TYPE));
      /* and the data */
      sp += mem_strcpy(mem_fn, Write, sp, argv[i]);
    }
  /* terminate argv array */
  mem_fn(Write, argv_addr + i*sizeof(SS_PTR_TYPE),
	 &null_ptr, sizeof(SS_PTR_TYPE));

  /* write envp pointer array and data */
  for (i = 0; envp[i]; i++)
    {
      /* write the envp pointer array entry */
      temp = SWAP_WORD(sp);
      mem_fn(Write, envp_addr + i*sizeof(SS_PTR_TYPE),
	     &temp, sizeof(SS_PTR_TYPE));
      /* and the data */
      sp += mem_strcpy(mem_fn, Write, sp, envp[i]);
    }
  /* terminate the envp array */
  mem_fn(Write, envp_addr + i*sizeof(SS_PTR_TYPE),
	       &null_ptr, sizeof(SS_PTR_TYPE));

  if (sp > ld_stack_base)
    fatal("environment overflow, increase SS_MAX_ENVIRON");

  debug("ld_text_base: 0x%08x  ld_text_size: 0x%08x",
	ld_text_base, ld_text_size);
  debug("ld_data_base: 0x%08x  ld_data_size: 0x%08x",
	ld_data_base, ld_data_size);
  debug("ld_stack_base: 0x%08x  ld_stack_size: 0x%08x",
	ld_stack_base, ld_stack_size);
  debug("ld_prog_entry: 0x%08x", ld_prog_entry);

}
