/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include "misc.h"
#include "ss.h"
#include "regs.h"
#include "memory.h"
#include "cache.h"
#include "loader.h"
#include "syscall.h"
#include "bpred.h"
#include "eventq.h"
#include "resource.h"
#include "sim.h"

/* track number of insn and refs */
static SS_COUNTER_TYPE num_insn = 0;
static SS_COUNTER_TYPE num_refs = 0;

static SS_COUNTER_TYPE num_loads = 0;
static SS_COUNTER_TYPE num_GPloads = 0;
static SS_COUNTER_TYPE num_SPloads = 0;
static SS_COUNTER_TYPE num_RRloads = 0;
static SS_COUNTER_TYPE num_GPstores = 0;
static SS_COUNTER_TYPE num_SPstores = 0;
static SS_COUNTER_TYPE num_RRstores = 0;

static SS_COUNTER_TYPE num_branches = 0;

static struct cache *icache = NULL;
static SS_COUNTER_TYPE icache_latency = 0;
static struct cache *dcache = NULL;
static SS_COUNTER_TYPE dcache_latency = 0;
static struct cache *tlb = NULL;

static struct bpred *pred = NULL;

static int issue_width = 4;

static int flush_on_syscalls = FALSE;

static int contiguous_fetch = TRUE;

static int perfect_memory = FALSE;

static int blocking_cache = FALSE;

/* mis-prediction cycle penalty */
static int num_cycle_MP = 2;

/* cycle counter */
static SS_TIME_TYPE cycle = 0;

/* block access latency */
static int blk_access_latency = 6;

/* functional unit resource pool */
struct res_pool *fu_pool = NULL;

/* functional unit resource configuration */
struct res_desc fu_config[] = {
  {
    "integer-ALU",
    4,
    0,
    {
      { IntALU, 1, 1 }
    }
  },
  {
    "integer-MULT/DIV",
    1,
    0,
    {
      { IntMULT, 3, 1 },
      { IntDIV, 20, 19 }
    }
  },
  {
    "memory-port",
    2,
    0,
    {
      { RdPort, 1, 1 },
      { WrPort, 1, 1 }
    }
  },
  {
    "FP-adder",
    4,
    0,
    {
      { FloatADD, 2, 1 },
      { FloatCMP, 2, 1 },
      { FloatCVT, 2, 1 }
    }
  },
  {
    "FP-MULT/DIV",
    1,
    0,
    {
      { FloatMULT, 4, 1 },
      { FloatDIV, 12, 12 },
      { FloatSQRT, 24, 24 }
    }
  },
};

static unsigned long
blk_access_fn(enum mem_cmd cmd, SS_ADDR_TYPE baddr, int bsize,
	      struct cache_blk *blk, SS_TIME_TYPE now)
{
  mem_access(cmd, baddr, &blk->data[0], bsize);
  return (cmd == Read) ? blk_access_latency : 0;
}

/* simulator-specific options */
char *sim_optstring = "fd:i:t:b:s:ypkm:j:w:";

void
sim_options(int argc, char **argv)
{
  char c;

  /* parse options */
  getopt_init();
  while ((c = getopt_next(argc, argv, sim_optstring)) != EOF)
    {
      switch (c) {
      case 'w':
	issue_width = atoi(getopt_arg);
	if (issue_width <= 0)
	  fatal("issue width must be positive and non-zero");
	if ((issue_width & (issue_width-1)) != 0)
	  fatal("issue width must be a power of two");
	break;
      case 'b':
	/* use BTB predictor */
	pred = bpred_create(BPred2bitBTB, atoi(getopt_arg));
	break;
      case 'j':
	num_cycle_MP = atoi(getopt_arg);
	if (num_cycle_MP < 1)
	  fatal("mis-prediction penalty must be at least 1 cycle");
	break;
      case 's':
	/* use static predictor */
	if (!strcmp(getopt_arg, "taken"))
	  pred = bpred_create(BPredTaken, 0);
	else if (!strcmp(getopt_arg, "nottaken"))
	  pred = bpred_create(BPredNotTaken, 0);
	else
	  fatal("cannot grok predictor type");
	break;
      case 'y':
	contiguous_fetch = FALSE;
	break;
      case 'i':
	{
	  char name[128], c;
	  int nsets, bsize, assoc;

	  /* use a I-cache */
	  if (sscanf(getopt_arg, "%[^:]:%d:%d:%d:%c",
		     name, &nsets, &bsize, &assoc, &c) != 5)
	    fatal("bad I-cache parms: <name>:<nsets>:<bsize>:<assoc>:<repl>");
	  icache =
	    cache_create(name, nsets, 2*bsize, /* balloc */TRUE, /* usize */0,
			 assoc, cache_char2policy(c), blk_access_fn,
			 /* hit latency */1);
	  info("** block size automatically doubled to %d bytes...", 2*bsize);
	}
	break;
      case 'd':
	{
	  char name[128], c;
	  int nsets, bsize, assoc;

	  /* use a D-cache */
	  if (sscanf(getopt_arg, "%[^:]:%d:%d:%d:%c",
		     name, &nsets, &bsize, &assoc, &c) != 5)
	    fatal("bad D-cache parms: <name>:<nsets>:<bsize>:<assoc>:<repl>");
	  dcache =
	    cache_create(name, nsets, bsize, /* balloc */TRUE, /* usize */0,
			 assoc, cache_char2policy(c), blk_access_fn,
			 /* hit latency */1);
	}
	break;
      case 'f':
	flush_on_syscalls = TRUE;
	break;
      case 'p':
	perfect_memory = TRUE;
	info("running with perfect (infinite B/W) memory...");
	break;
      case 'k':
	blocking_cache = TRUE;
	info("running with a blocking cache interface...");
	break;
      case 'm':
	if (icache || dcache)
	  fatal("specify `-m' before `-i' or `-d'");
	blk_access_latency = atoi(getopt_arg);
	break;
      }
    }
}

void
sim_config(FILE *stream)
{
  info("issuing up to %d instructions per cycle", issue_width);
  if (icache)
    cache_config(icache, stream);
  if (dcache)
    cache_config(dcache, stream);
  info("cache miss latency = %d cycles", blk_access_latency);
  info("mis-prediction penalty = %d cycles", num_cycle_MP);
  if (flush_on_syscalls)
    info("flushing D-cache on system calls...");
  else
    info("running system calls through warm D-cache...");
  if (tlb)
    cache_config(tlb, stream);
  if (pred)
    bpred_config(pred, stream);
  else
    info("using no predictor...");
  if (contiguous_fetch)
    info("breaking I-fetch on branches...");
  else
    info("*not* breaking I-fetch on branches...");
}

/* set to non-zero when simulator should dump statistics */
int sim_dump_stats = FALSE;

void
sim_stats(FILE *stream)
{
  extern time_t start_time;
  time_t elapsed = MAX(time((time_t *)NULL) - start_time, 1);

  info("sim: executed %.0f instructions", (double)num_insn);
  info("sim: executed %.0f cycles", (double)cycle);
  info("sim: IPC: %f  CPI: %f",
       (double)num_insn/(double)cycle, (double)cycle/(double)num_insn);
  info("sim: accessed (data) memory %.0f times", (double)num_refs);
  info("sim: simulation time: %s (%f insts/sec)",
       elapsed_time(elapsed), (double)num_insn/(double)elapsed);
  info("sim: executed %.0f branches (%f insts/branch",
       (double)num_branches, (double)num_insn/(double)num_branches);
  if (icache)
    cache_stats(icache, stream);
  info("icache: I-cache access latency = %f",
       (double)icache_latency/(double)num_insn);
  if (dcache)
    cache_stats(dcache, stream);
  info("dcache: D-cache access latency = %f",
       (double)dcache_latency/(double)num_loads);
  if (tlb)
    cache_stats(tlb, stream);
  if (pred)
    bpred_stats(pred, stream);
}

static struct execute_rec {
  SS_INST_TYPE IR;			/* inst register */
  SS_ADDR_TYPE PC, pred_PC;		/* current PC, predicted next PC */
  struct res_template *fu;		/* functional unit allocated to inst */
  int out1, out2;			/* instruction outputs */
} *execute_data;
static int execute_num;			/* num entries in execute latches */
BITMAP_TYPE(SS_TOTAL_REGS, busy);	/* architected registers busy bits */
static unsigned long ifetch_block = 0;

void
execute_squash(void)
{
  execute_num = 0;
}

void
execute_init(void)
{
  execute_data =
    (struct execute_rec *)calloc(issue_width, sizeof(struct execute_rec));
  if (!execute_data)
    fatal("out of virtual memory");
  execute_squash();
  BITMAP_CLEAR_MAP(busy, BITMAP_SIZE(SS_TOTAL_REGS));
}

/* configure the execution engine */
#define SET_NPC(EXPR)		(next_PC = (EXPR))
#define CPC			(execute_data[i].PC)

#define GPR(N)			(regs_R[N])
/* Doug added this */
#define GPR_D(N)		(regs_R[N >> 1])
/* *************** */
#define SET_GPR(N,EXPR)		(regs_R[N] = (EXPR))

#define FPR_L(N)		(regs_F.l[(N)])
#define SET_FPR_L(N,EXPR)	(regs_F.l[(N)] = (EXPR))
#define FPR_F(N)		(regs_F.f[(N)])
#define SET_FPR_F(N,EXPR)	(regs_F.f[(N)] = (EXPR))
#define FPR_D(N)		(regs_F.d[(N) >> 1])
#define SET_FPR_D(N,EXPR)	(regs_F.d[(N) >> 1] = (EXPR))

#define SET_HI(EXPR)		(regs_HI = (EXPR))
#define HI			(regs_HI)
#define SET_LO(EXPR)		(regs_LO = (EXPR))
#define LO			(regs_LO)
#define FCC			(regs_FCC)
#define SET_FCC(EXPR)		(regs_FCC = (EXPR))

#define __READ_CACHE(DST_T, SRC_T, SRC)					\
  ({ SRC_T temp; num_refs++; num_loads++;				\
     if (dcache)							\
       (cachelat = cache_access(dcache, Read, (SRC),			\
				&temp, sizeof(SRC_T), cycle, NULL, NULL),	\
	dcache_latency += cachelat, NULL);					\
     else								\
       (cachelat = 0, mem_access(Read, (SRC), &temp, sizeof(SRC_T)));	\
     ((unsigned long)((DST_T)temp)); })

#define READ_WORD(SRC)							\
  __READ_CACHE(unsigned long, unsigned long, (SRC))

#define READ_UNSIGNED_HALF(SRC)						\
  __READ_CACHE(unsigned long, unsigned short, (SRC))

#define READ_SIGNED_HALF(SRC)						\
  __READ_CACHE(signed long, signed short, (SRC))

#define READ_UNSIGNED_BYTE(SRC)						\
  __READ_CACHE(unsigned long, unsigned char, (SRC))

#define READ_SIGNED_BYTE(SRC)						\
  __READ_CACHE(signed long, signed char, (SRC))

#define __WRITE_CACHE(SRC, DST_T, DST)					\
  ({ DST_T temp = (DST_T)((unsigned long)(SRC));			\
     num_refs++;							\
     if (dcache)							\
       cachelat = cache_access(dcache, Write, (DST),			\
			       &temp, sizeof(DST_T), cycle, NULL, NULL);	\
     else								\
       (cachelat = 0, mem_access(Write, (DST), &temp, sizeof(DST_T))); })

#define WRITE_WORD(SRC, DST)						\
  __WRITE_CACHE((SRC), unsigned long, (DST))

#define WRITE_HALF(SRC, DST)						\
  __WRITE_CACHE((SRC), unsigned short, (DST))

#define WRITE_BYTE(SRC,DST)						\
  __WRITE_CACHE((SRC), unsigned char, (DST))

static void
dcache_access_fn(enum mem_cmd cmd, SS_ADDR_TYPE addr, void *p, int nbytes)
{
  if (dcache)
    cache_access(dcache, cmd, addr, p, nbytes, cycle, NULL, NULL);
  else
    mem_access(cmd, addr, p, nbytes);
}

#define SYSCALL(INST)							\
  (flush_on_syscalls							\
   ? ((dcache ? cache_flush(dcache, 0) : 0),				\
      ss_syscall(mem_access, INST))					\
   : (ss_syscall(dcache_access_fn, INST)))

#define DEFFU(FU,DESC)
#define DEFICLASS(ICLASS,DESC)
#define DEFINST(OP,MSK,NAME,OPFORM,RES,CLASS,O1,O2,I1,I2,I3,EXPR)
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)
#define CONNECT(OP)
#define IMPL
#include "ss.def"
#undef DEFFU
#undef DEFICLASS
#undef DEFINST
#undef DEFLINK
#undef CONNECT
#undef IMPL

void
execute(void)
{
  int i, cachelat;
  SS_INST_TYPE inst;
  enum ss_opcode op;
  SS_ADDR_TYPE next_PC;
  struct res_template *fu;
  static void idecode_squash(void);
  static void ifetch_squash(void);

  for (i=0; i<execute_num; i++)
    {
      /* maintain $r0 semantics */
      regs_R[0] = 0;

      inst = execute_data[i].IR;
      op = SS_OPCODE(inst);

      assert(!(SS_OP_FLAGS(op) & F_TRAP) || execute_num == 1);

      /* compute default next PC */
      next_PC = execute_data[i].PC + sizeof(SS_INST_TYPE);

      /* compute default extra cache latency */
      cachelat = 0;

      /* execute the instruction */
      switch (op) {
#define DEFFU(FU,DESC)
#define DEFICLASS(ICLASS,DESC)
#define DEFINST(OP,MSK,NAME,OPFORM,RES,CLASS,O1,O2,I1,I2,I3,EXPR)	\
      case OP:								\
	EXPR; num_insn++;						\
	break;
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)					\
      case OP:								\
	panic("tried to execute a linking opcode, %d", OP);
#define CONNECT(OP)
#include "ss.def"
#undef DEFFU
#undef DEFICLASS
#undef DEFINST
#undef DEFLINK
#undef CONNECT
      default:
	panic("bogus opcode `%d' encountered", op);
      }

      if (SS_OP_FLAGS(op) & F_LOAD)
	{
	  if (SS_OP_FLAGS(op) & F_DISP)
	    {
	      if ((BS) == Rsp)
		num_SPloads++;
	      else if ((BS) == Rsp)
		num_GPloads++;
	    }
	  else if (SS_OP_FLAGS(op) & F_RR)
	    {
	      num_RRloads++;
	      if ((RD) == Rsp || (BS) == Rsp)
		num_SPloads++;
	    }
	  else if (SS_OP_FLAGS(op) & F_DIRECT)
	    {
	      /* FIXME: */
	      ;
	    }
	  else
	    abort();
	}

      if (SS_OP_FLAGS(op) & F_STORE)
	{
	  if (SS_OP_FLAGS(op) & F_DISP)
	    {
	      if ((BS) == Rsp)
		num_SPstores++;
	      else if ((BS) == Rgp)
		num_GPstores++;
	    }
	  else if (SS_OP_FLAGS(op) & F_RR)
	    {
	      num_RRstores++;
	      if ((RD) == Rsp || (BS) == Rsp)
		num_SPstores++;
	    }
	  else if (SS_OP_FLAGS(op) & F_DIRECT)
	    {
	      /* FIXME: */
	      ;
	    }
	  else
	    abort();
	}

      if (SS_OP_FLAGS(op) & F_CTRL)
	{
	  num_branches++;

	  /* update the branch target buffer state/stats */
	  if (pred)
	    bpred_update(pred,
			 execute_data[i].PC,
			 next_PC,
			 /* was the branch actually taken */
			 next_PC != (execute_data[i].PC+sizeof(SS_INST_TYPE)),
			 /* good pred? */execute_data[i].pred_PC == next_PC);
	}

      /* this inst will execute, schedule release events */
      if ((fu = execute_data[i].fu))
	{
	  int oplat = fu->oplat + cachelat;

	  /* manually perform release event if latency is one cycle */
	  if (fu->issuelat <= 1)
	    {
	      assert(fu->master->busy);
	      fu->master->busy = FALSE;
	    }
	  else
	    eventq_queue_setflag(cycle+(fu->issuelat-1),
				 &fu->master->busy, FALSE);
	  if (oplat <= 1)
	    {
	      if (execute_data[i].out1)
		{
		  BITMAP_CLEAR(busy, BITMAP_SIZE(SS_TOTAL_REGS),
			       execute_data[i].out1);
		}
	      if (execute_data[i].out2)
		{
		  BITMAP_CLEAR(busy, BITMAP_SIZE(SS_TOTAL_REGS),
			       execute_data[i].out2);
		}
	    }
	  else
	    {
	      if (execute_data[i].out1)
		{
		  eventq_queue_clearbit(cycle+(oplat-1), busy,
					BITMAP_SIZE(SS_TOTAL_REGS),
					execute_data[i].out1);
		}
	      if (execute_data[i].out2)
		{
		  eventq_queue_clearbit(cycle+(oplat-1), busy,
					BITMAP_SIZE(SS_TOTAL_REGS),
					execute_data[i].out2);
		}
	    }
	}
      else
	{
	  /* shouldn't be any busy bit events */
	  assert(!execute_data[i].out1 && !execute_data[i].out2);
	}

      /* was a branch mis-predicted? if so, recover */
      if (execute_data[i].pred_PC != next_PC)
	{
	  /* reset any previously set busy bits, release functional units */
	  for (i++; i<execute_num; i++)
	    {
	      /* undo reservation and busy bit actions by executing them */
	      if ((fu = execute_data[i].fu))
		{
		  assert(fu->master->busy);
		  fu->master->busy = FALSE;

		  if (execute_data[i].out1)
		    {
		      BITMAP_CLEAR(busy, BITMAP_SIZE(SS_TOTAL_REGS),
				   execute_data[i].out1);
		    }
		  if (execute_data[i].out2)
		    {
		      BITMAP_CLEAR(busy, BITMAP_SIZE(SS_TOTAL_REGS),
				   execute_data[i].out2);
		    }
		}
	      else
		{
		  /* shouldn't be any busy bit events */
		  assert(!execute_data[i].out1 && !execute_data[i].out2);
		}
	    }

	  /* squash */
	  regs_PC = next_PC;
	  execute_squash();
	  idecode_squash();
	  ifetch_squash();

	  /* mis-prediction latency */
	  ifetch_block += (num_cycle_MP-1);

	  /* bail */
	  return;
	}
    }

  /* all instruction processed */
  execute_num = 0;
}

static struct idecode_rec {
  SS_INST_TYPE IR;			/* inst register */
  SS_ADDR_TYPE PC, pred_PC;		/* current PC, predicted next PC */
} *idecode_data;
static int idecode_num;			/* num entries in idecode queue */
static int idecode_tail, idecode_head;	/* head and tail pointer of queue */

static void
idecode_squash(void)
{
  idecode_num = 0;
  idecode_tail = idecode_head = 0;
}

static void
idecode_init(void)
{
  idecode_data =
    (struct idecode_rec *)calloc(issue_width, sizeof(struct idecode_rec));
  if (!idecode_data)
    fatal("out of virtual memory");
  idecode_squash();
}

#undef GPR(N)
#undef GPR_D(N)
#undef CGPR(N)
#undef FPR_L(N)
#undef FPR_F(N)
#undef FPR_D(N)
#undef HI
#undef LO
#undef FCC
#undef MEM
#undef CTRL

/* $r0 must be (0) NA, which it is */
#define GPR(N)		(N)
/* Doug added this */
#define GPR_D(N)	((N)&~1)
/* *************** */
/* no conditional dependencies */
#define CGPR(N)		(0)
/* all FP reg dependencies resolve to a dependence on the pair of regs */
#define FPR_L(N)	(((N)+32)&~1)
#define FPR_F(N)	(((N)+32)&~1)
#define FPR_D(N)	(((N)+32)&~1)
/* special regs */
#define HI		(0+32+32)
#define LO		(1+32+32)
#define FCC		(2+32+32)
/* dependencies not tracked, therefore link to $r0 */
#define MEM		(0)
#define CTRL		(0)

/* build the input dependence mask */
#define I_GPR(N)						\
  BITMAP_SET(inputs, BITMAP_SIZE(SS_TOTAL_REGS), GPR(N))
/* Doug added this */
#define I_GPR_D(N)						\
  BITMAP_SET(inputs, BITMAP_SIZE(SS_TOTAL_REGS), GPR_D(N))
/* *************** */
#define I_FPR_L(N)						\
  BITMAP_SET(inputs, BITMAP_SIZE(SS_TOTAL_REGS), FPR_L(N))
#define I_FPR_F(N)						\
  BITMAP_SET(inputs, BITMAP_SIZE(SS_TOTAL_REGS), FPR_F(N))
/* Doug added this, changing FPR_F to FPR_D */
#define I_FPR_D(N)						\
  BITMAP_SET(inputs, BITMAP_SIZE(SS_TOTAL_REGS), FPR_D(N))
/* *************** */
#define I_HI							\
  BITMAP_SET(inputs, BITMAP_SIZE(SS_TOTAL_REGS), HI)
#define I_LO							\
  BITMAP_SET(inputs, BITMAP_SIZE(SS_TOTAL_REGS), LO)
#define I_FCC							\
  BITMAP_SET(inputs, BITMAP_SIZE(SS_TOTAL_REGS), FCC)
#define I_MEM		/* NOP: was RSSetBit(inputs, MEM) */
#define I_CTRL		/* NOP: was RSSetBit(inputs, CTRL) */
#define I_NA

static void
idecode(void)
{
  int i, ndecode = idecode_num;
  SS_INST_TYPE inst;
  enum ss_opcode op;
  struct res_template *fu;
  int out1, out2;
  BITMAP_TYPE(SS_TOTAL_REGS, inputs);
  /* a load and store cannot issue in parallel */
  int issued_load = FALSE, issued_store = FALSE;

  for (i=0; i<ndecode; i++)
    {
      inst = idecode_data[idecode_head].IR;
      op = SS_OPCODE(inst);

      fu = NULL;
      out1 = 0; out2 = 0;
      BITMAP_CLEAR_MAP(inputs, BITMAP_SIZE(SS_TOTAL_REGS));

      switch (op) {
#define DEFFU(FU,DESC)
#define DEFICLASS(ICLASS,DESC)
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR)	\
      case OP:								\
	/* output names */						\
	out1 = O1;  out2 = O2;						\
	/* build input dependence mask */				\
	I_##I1; I_##I2; I_##I3;						\
	break;
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)					\
    case OP:								\
	panic("tried to get dependency info for a linking opcode, %d", OP);
#define CONNECT(OP)
#include "ss.def"
#undef DEFFU
#undef DEFICLASS
#undef DEFINST
#undef DEFLINK
#undef CONNECT
      default:
	/* speculatively, we may (try to) decode a bogus opcode */
	op = NA;
	inst = SS_NOP_INST;
      }

      /* $r0 is a zero source/sink register */
      BITMAP_CLEAR(inputs, BITMAP_SIZE(SS_TOTAL_REGS), 0);

      if (SS_OP_FLAGS(op) & F_TRAP)
	{
	  /* system call instruction, this needs to execute all alone,
	     the stage of execute is unimportant as all SYSCALL insts will
	     be preceded by a LI R2,0x... instruction, which will always
	     complete in the EX stage */
	  if (i == 0)
	    {
	      assert(execute_num == 0);

	      /* commit this instruction to the next stage */
	      execute_data[execute_num].IR = idecode_data[idecode_head].IR;
	      execute_data[execute_num].PC = idecode_data[idecode_head].PC;
	      execute_data[execute_num].pred_PC =
		idecode_data[idecode_head].pred_PC;
	      execute_data[execute_num].fu = fu;
	      execute_data[execute_num].out1 = out1;
	      execute_data[execute_num].out2 = out2;
	      execute_num++;

	      idecode_head = (idecode_head + 1) & (issue_width - 1);
	      idecode_num--;
	      return;
	    }
	  else
	    {
	      /* it will execute in the next cycle, otherwise, in either case
		 there is something for the EX stage to do in the next cycle */
	      return;
	    }
	}

      /* enforce basic pipeline issue constraints */

      /* stall on RAW hazards */
      if (!BITMAP_DISJOINT_P(busy, inputs, BITMAP_SIZE(SS_TOTAL_REGS)))
	{
	  /* must wait until input register is ready */
	  return;
	}

      if (!perfect_memory)
	{
	  if (((SS_OP_FLAGS(op) & F_LOAD) && issued_store)
	      || ((SS_OP_FLAGS(op) & F_STORE) &&
		  (issued_load || issued_store)))
	    {
	      /* only one store at a time */
	      return;
	    }

	  /* sanity check */
	  assert(!issued_load || !issued_store);
	}

      /* must use a functional unit to create a result */
      assert(SS_OP_FUCLASS(op) != NA || (out1 == NA && out2 == NA));

      /* get the necessary resources */
      if (SS_OP_FUCLASS(op) != NA)
	{
	  if ((fu = res_get(fu_pool, SS_OP_FUCLASS(op))))
	    {
	      /* got a resource, aquire resource and set busy bits;
	         NOTE: release events are not scheduled here, they are
		 instead scheduled in EXECUTE(), this results in two
		 optimization:
		   1) single cycle events do not go into the event queue
		   2) squashed insts never actually schedule an event */

	      /* aquire functional unit */
	      assert(!fu->master->busy);
	      fu->master->busy = TRUE;

	      /* set result busy bit(s) */
	      if (out1)
		{
		  BITMAP_SET(busy, BITMAP_SIZE(SS_TOTAL_REGS), out1);
		}
	      if (out2)
		{
		  BITMAP_SET(busy, BITMAP_SIZE(SS_TOTAL_REGS), out2);
		}
	    }
	  else
	    {
	      /* no resource available, inorder issue, so bail */
	      return;
	    }
	}
      /* else, no resource required (therefore no outputs) */

      /* inst issued, result registers are now pending results */

      if (SS_OP_FLAGS(op) & F_LOAD)
	issued_load = TRUE;
      else if (SS_OP_FLAGS(op) & F_STORE)
	issued_store = TRUE;

      /* commit this instruction to the next stage */
      execute_data[execute_num].IR = idecode_data[idecode_head].IR;
      execute_data[execute_num].PC = idecode_data[idecode_head].PC;
      execute_data[execute_num].pred_PC = idecode_data[idecode_head].pred_PC;
      execute_data[execute_num].fu = fu;
      execute_data[execute_num].out1 = out1;
      execute_data[execute_num].out2 = out2;
      execute_num++;

      idecode_head = (idecode_head + 1) & (issue_width - 1);
      idecode_num--;
    }
  assert(idecode_num == 0);
}

static void
ifetch_squash(void)
{
  ifetch_block = 0;
}

static void
ifetch_init(void)
{
  ifetch_block = 0;
}

/* fetch up to NUMISSUE instructions, the fetched instructions are
   placed into the circular queue owned by the decode stage; when the
   decode queue is full fetching terminates; fetching also terminates
   when the BTB predicts a jump, since the instructions at the new
   address will not be available until the next cycle */
static void
ifetch(void)
{
  int i, lat, done = FALSE;
  SS_ADDR_TYPE target, inst_PC;
  SS_INST_TYPE inst;

  /* handle any untolerated I-cache latencies */
  if (ifetch_block)
    {
      /* block on an I-cache miss */
      ifetch_block--;
      return;
    }

  for (i=0;
       (i < issue_width)		/* fetch up to NUMISSUE insts */
       && (idecode_num < issue_width)	/* fetch until idecode() stage full */
       && !done;
       i++)
    {
      /* fetch instruction */
      inst_PC = regs_PC;
      if (icache)
	{
	  lat = cache_access(icache, Read, regs_PC,
			     &inst, sizeof(SS_INST_TYPE), cycle, NULL, NULL);
	  if (lat != 1)
	    {
	      /* I-cache miss */

	      /* may not be the case due to squashes */
	      /* assert(lat == blk_access_latency); */

	      ifetch_block += lat - 1;
	      icache_latency += lat - 1;
	      break;
	    }
	  /* else: cache hit */
	}
      else
	mem_access(Read, regs_PC, &inst, sizeof(SS_INST_TYPE));

      /* have a valid inst, here */

      /* possibly use the BTB target */
      if (pred && (target = bpred_lookup(pred, regs_PC, 0)))
	{
	  /* predicted taken */
	  regs_PC = target;
	  /* discontinuous stream, so stop fetching if needed */
	  done = contiguous_fetch;
	}
      else
	{
	  /* predicted not taken, continue fetching linearly */
	  regs_PC += sizeof(SS_INST_TYPE);
	}

      /* commit this instruction to the next stage */
      idecode_data[idecode_tail].IR = inst;
      idecode_data[idecode_tail].PC = inst_PC;
      idecode_data[idecode_tail].pred_PC = regs_PC;
      idecode_tail = (idecode_tail + 1) & (issue_width - 1);
      idecode_num++;
    }
}

static char *reg_names[] = {
  "r0", "r1", "r2", "r3", "r4", "r5", "r6", "r7",
  "r8", "r9", "r10", "r11", "r12", "r13", "r14", "r15",
  "r16", "r17", "r18", "r19", "r20", "r21", "r22", "r23",
  "r24", "r25", "r26", "r27", "r28", "r29", "r30", "r31",
  "f0", "f1", "f2", "f3", "f4", "f5", "f6", "f7",
  "f8", "f9", "f10", "f11", "f12", "f13", "f14", "f15",
  "f16", "f17", "f18", "f19", "f20", "f21", "f22", "f23",
  "f24", "f25", "f26", "f27", "f28", "f29", "f30", "f31",
  "HI", "LO", "FCC", "MEM", "CTRL",
};

static void
dump_reg_map(BITMAP_ENT_TYPE *bmap, FILE *stream)
{
  int i;

  for (i=0; i<SS_TOTAL_REGS; i++)
    {
      if (BITMAP_SET_P(bmap, BITMAP_SIZE(SS_TOTAL_REGS), i))
	fprintf(stream, "%s ", reg_names[i]);
    }
}

static void
dump_pipe_state(FILE *stream)
{
  int i, j;

  if (!stream)
    stream = stderr;

  fprintf(stream, ">>>> Pipe state <<<<\n");
  fprintf(stream, ">> cycle: %.0f\n", (double)cycle);
  fprintf(stream, ">> "); eventq_dump(stream); fprintf(stream, "\n");
  fprintf(stream, ">> busy bits: "); dump_reg_map(busy, stream);
  fprintf(stream, "\n");
  fprintf(stream, ">> execute:\n");
  for (i=0; i<execute_num; i++)
    {
      fprintf(stream, ">>   PC: 0x%08lx  (P)NPC: 0x%08lx  IR: ",
	      execute_data[i].PC, execute_data[i].pred_PC);
      ss_print_insn(execute_data[i].IR, execute_data[i].PC, stream);
      fprintf(stream, "\n");
      fprintf(stream, ">>   OP: %s  FU: %s  flags: 0x%08lx\n",
	      SS_OP_NAME(SS_OPCODE(execute_data[i].IR)),
	      SS_FU_NAME(SS_OP_FUCLASS(SS_OPCODE(execute_data[i].IR))),
	      SS_OP_FLAGS(SS_OPCODE(execute_data[i].IR)));
      fprintf(stream, ">>   resv FU: %s(%d)  out1: %d  out2: %d\n",
	      execute_data[i].fu ? execute_data[i].fu->master->name : "(none)",
	      execute_data[i].fu ? execute_data[i].fu->master->busy : 0,
	      execute_data[i].out1, execute_data[i].out2);
    }
  fprintf(stream, ">> idecode:\n");
  for (i=0,j=idecode_head; i<idecode_num; i++,j=((j+1) & (issue_width-1)))
    {
      fprintf(stream, ">>   PC: 0x%08lx  (P)NPC: 0x%08lx  IR: ",
	      idecode_data[j].PC, idecode_data[j].pred_PC);
      ss_print_insn(idecode_data[j].IR, idecode_data[j].PC, stream);
      fprintf(stream, "\n");
      fprintf(stream, ">>   OP: %s  FU: %s  flags: 0x%08lx\n",
	      SS_OP_NAME(SS_OPCODE(idecode_data[j].IR)),
	      SS_FU_NAME(SS_OP_FUCLASS(SS_OPCODE(idecode_data[j].IR))),
	      SS_OP_FLAGS(SS_OPCODE(idecode_data[j].IR)));
    }
  fprintf(stream, ">> ifetch:\n");
  fprintf(stream, ">>   PC: 0x%08lx  block count: %ld\n",
	  regs_PC, ifetch_block);
}
/* exit when this becomes non-zero */
int sim_exit_now = FALSE;

/* longjmp here when simulation is completed */
jmp_buf sim_exit_buf;

void
sim_main(void)
{
  info("starting in-order issue simulation");

  num_insn = 0;
  num_refs = 0;

  /* decode all instructions */
  {
    SS_ADDR_TYPE addr;
    SS_INST_TYPE inst;

    if (OP_MAX > 255)
      fatal("cannot do fast decoding, too many opcodes");

    fprintf(stderr, "decoding text segment...");
    for (addr=ld_text_base;
	 addr < (ld_text_base+ld_text_size);
	 addr += SS_INST_SIZE)
      {
	inst = __UNCHK_MEM_ACCESS(SS_INST_TYPE, addr);
	inst.a = (inst.a & ~0xff) | (unsigned long)SS_OP_ENUM(SS_OPCODE(inst));
	__UNCHK_MEM_ACCESS(SS_INST_TYPE, addr) = inst;
      }
    fprintf(stderr, "done.\n");
  }

  /* various initializations */
  fu_pool = res_create_pool("fu-pool", fu_config, N_ELT(fu_config));
  eventq_init(/* max events: 2-busy + fu release */3*N_ELT(fu_config));
  execute_init();
  idecode_init();
  ifetch_init();
  regs_PC = ld_prog_entry;

  for (;;)
    {
      if (verbose)
	dump_pipe_state(stderr);

      /* commit any writeback events for this cycle */
      /* FIXME: could limit writeback bandwidth here */
      eventq_service_events(cycle);

      /* execute instructions */
      execute();

      /* decode fetched instructions */
      idecode();

      /* fetch instructions for next cycle */
      ifetch();

      /* go to next cycle */
      cycle++;
    }
}
